import unittest
import navipy.processing.mcode as mcode
import pandas as pd
# import matplotlib.pyplot as plt
import numpy as np


def Scale(data, oldmin, oldmax, mini, maxi, ran):
    """Scales all values in data into the range mini, maxi
    Input:
        - data: values to be scaled
        - mini: minimum of new range
        - maxi: maximum of new range
        - ran: range of new values (ran=maxi-mini)

    Output:
        - data: the scaled data"""
    data = (data-oldmin)/oldmax
    dmax = np.max(data)
    dmin = np.min(data)
    scaling = ran/(dmax-dmin)
    data = (data-dmin)*scaling
    data = data+mini
    return data


class TestCase(unittest.TestCase):
    def setUp(self):
        """sets up a distance array that is used in some tests,
           as well as the velocity and viewing direction frames
        """
        distance = np.array([[2.60009956,  2.60009956,  2.60009956],
                             [2.60009956,  2.60009956,  2.60009956],
                             [2.60009956,  2.60009956,  2.60009956]])

        self.scene = np.zeros((3, 3, 4, 1))
        self.scene[:, :, 0, 0] = np.array(distance).copy()
        self.scene[:, :, 1, 0] = np.array(distance).copy()
        self.scene[:, :, 2, 0] = np.array(distance).copy()
        self.scene[:, :, 3, 0] = np.array(distance).copy()

        self.velocity = pd.Series(index=['x', 'y', 'z',
                                         'alpha_0', 'alpha_1', 'alpha_2',
                                         'dx', 'dy', 'dz',
                                         'dalpha_0', 'dalpha_1', 'dalpha_2'])

        self.elevation = np.arange(-np.pi/2, 4*(np.pi/360)-np.pi/2,
                                   2*(np.pi/360))
        self.azimuth = np.arange(0, 4*(np.pi/360)+2*(np.pi/360), 2*(np.pi/360))

        self.viewing_directions = np.zeros((3, 2))
        self.viewing_directions[:, 0] = self.elevation
        self.viewing_directions[:, 1] = self.azimuth

    def test_rotation_not_too_strong(self):
        """
        this test checks that a Value Error is throught,
        if the change in rotation is more than pi/2.
        Thoughs, if the velocity of yaw, pitch and roll is
        is smaller than pi/2
        """
        # dyaw is too big
        positions = np.array([[-20, -20, 2.6, 1.57079633, 0, 0],
                              [-19.8, -20, 2.6,
                               1.57079633 + np.pi/2 + 0.00001, 0, 0]])

        x = positions[0, 0]
        y = positions[0, 1]
        z = positions[0, 2]
        yaw = positions[0, 3]
        pitch = positions[0, 4]
        roll = positions[0, 5]

        dx = positions[1, 0] - positions[0, 0]
        dy = positions[1, 1] - positions[0, 1]
        dz = positions[1, 2] - positions[0, 2]
        dyaw = positions[1, 3] - positions[0, 3]
        dpitch = positions[1, 4] - positions[0, 4]
        droll = positions[1, 5] - positions[0, 5]

        self.velocity.x = x
        self.velocity.y = y
        self.velocity.z = z
        self.velocity.alpha_0 = yaw
        self.velocity.alpha_1 = pitch
        self.velocity.alpha_2 = roll
        self.velocity.dx = dx
        self.velocity.dy = dy
        self.velocity.dz = dz
        self.velocity.dalpha_0 = dyaw
        self.velocity.dalpha_1 = dpitch
        self.velocity.dalpha_2 = droll

        with self.assertRaises(ValueError):
            rof, hof, vof = mcode.optic_flow(self.scene,
                                             self.viewing_directions,
                                             self.velocity)

        # dpitch is too big
        positions = np.array([[-20, -20, 2.6, 1.57079633, 0, 0],
                              [-19.8, -20, 2.6,
                               1.57079633, 0 + np.pi/2 + 0.0001, 0]])

        x = positions[0, 0]
        y = positions[0, 1]
        z = positions[0, 2]
        yaw = positions[0, 3]
        pitch = positions[0, 4]
        roll = positions[0, 5]

        dx = positions[1, 0] - positions[0, 0]
        dy = positions[1, 1] - positions[0, 1]
        dz = positions[1, 2] - positions[0, 2]
        dyaw = positions[1, 3] - positions[0, 3]
        dpitch = positions[1, 4] - positions[0, 4]
        droll = positions[1, 5] - positions[0, 5]

        self.velocity.x = x
        self.velocity.y = y
        self.velocity.z = z
        self.velocity.alpha_0 = yaw
        self.velocity.alpha_1 = pitch
        self.velocity.alpha_2 = roll
        self.velocity.dx = dx
        self.velocity.dy = dy
        self.velocity.dz = dz
        self.velocity.dalpha_0 = dyaw
        self.velocity.dalpha_1 = dpitch
        self.velocity.dalpha_2 = droll

        with self.assertRaises(ValueError):
            rof, hof, vof = mcode.optic_flow(self.scene,
                                             self.viewing_directions,
                                             self.velocity)

        # droll is too big
        positions = np.array([[-20, -20, 2.6, 1.57079633, 0, 0],
                              [-19.8, -20, 2.6, 1.57079633, 0,
                               0 + np.pi/2 + 0.0001]])

        x = positions[0, 0]
        y = positions[0, 1]
        z = positions[0, 2]
        yaw = positions[0, 3]
        pitch = positions[0, 4]
        roll = positions[0, 5]

        dx = positions[1, 0] - positions[0, 0]
        dy = positions[1, 1] - positions[0, 1]
        dz = positions[1, 2] - positions[0, 2]
        dyaw = positions[1, 3] - positions[0, 3]
        dpitch = positions[1, 4] - positions[0, 4]
        droll = positions[1, 5] - positions[0, 5]

        self.velocity.x = x
        self.velocity.y = y
        self.velocity.z = z
        self.velocity.alpha_0 = yaw
        self.velocity.alpha_1 = pitch
        self.velocity.alpha_2 = roll
        self.velocity.dx = dx
        self.velocity.dy = dy
        self.velocity.dz = dz
        self.velocity.dalpha_0 = dyaw
        self.velocity.dalpha_1 = dpitch
        self.velocity.dalpha_2 = droll

        with self.assertRaises(ValueError):
            rof, hof, vof = mcode.optic_flow(self.scene,
                                             self.viewing_directions,
                                             self.velocity)

    def test_only_x(self):
        """
        this test checks for the correct response if for example
        the bee moves only in x-direction keeping the other
        parameters (y,z,yaw,pitch,roll) constant
        """
        positions = np.array([[-20, -20, 2.6, 1.57079633, 0, 0],
                              [-19.8, -20, 2.6, 1.57079633, 0, 0]])

        x = positions[0, 0]
        y = positions[0, 1]
        z = positions[0, 2]
        yaw = positions[0, 3]
        pitch = positions[0, 4]
        roll = positions[0, 5]

        dx = positions[1, 0] - positions[0, 0]
        dy = positions[1, 1] - positions[0, 1]
        dz = positions[1, 2] - positions[0, 2]
        dyaw = positions[1, 3] - positions[0, 3]
        dpitch = positions[1, 4] - positions[0, 4]
        droll = positions[1, 5] - positions[0, 5]

        self.velocity.x = x
        self.velocity.y = y
        self.velocity.z = z
        self.velocity.alpha_0 = yaw
        self.velocity.alpha_1 = pitch
        self.velocity.alpha_2 = roll
        self.velocity.dx = dx
        self.velocity.dy = dy
        self.velocity.dz = dz
        self.velocity.dalpha_0 = dyaw
        self.velocity.dalpha_1 = dpitch
        self.velocity.dalpha_2 = droll

        rof, hof, vof = mcode.optic_flow(self.scene,
                                         self.viewing_directions,
                                         self.velocity)

        testrof = [[0.00000000e+00, 0.00000000e+00, 0.00000000e+00],
                   [-8.07793567e-28, 0.00000000e+00, 6.77626358e-21],
                   [-1.61558713e-27, 0.00000000e+00, 0.00000000e+00]]

        testhof = [[0.07692013, 0.07690842, 0.07687327],
                   [0.07692013, 0.07690842, 0.07687327],
                   [0.07692013, 0.07690842,  0.07687327]]

        testvof = [[2.46536979e-10, 1.34244164e-03, 2.68447412e-03],
                   [2.46499430e-10, 1.34223718e-03, 2.68406526e-03],
                   [2.46386795e-10, 1.34162387e-03, 2.68283881e-03]]

        assert np.all(np.isclose(rof, testrof))
        assert np.all(np.isclose(hof, testhof))
        assert np.all(np.isclose(vof, testvof))

    def test_only_yaw(self):
        """
        this test checks for the correct response if for example
        the bee rotates only around the yaw axis keeping the other
        parameters (x,y,z,pitch,roll) constant
        """
        # yaw only everything zero
        # only vertical should change, horizontal stays same
        positions = np.array([[-20, -20, 2.6, 1.57079633, 0, 0],
                              [-20, -20, 2.6, 1.90079633, 0, 0]])

        x = positions[0, 0]
        y = positions[0, 1]
        z = positions[0, 2]
        yaw = positions[0, 3]
        pitch = positions[0, 4]
        roll = positions[0, 5]

        dx = positions[1, 0]-positions[0, 0]
        dy = positions[1, 1]-positions[0, 1]
        dz = positions[1, 2]-positions[0, 2]
        dyaw = positions[1, 3]-positions[0, 3]
        dpitch = positions[1, 4]-positions[0, 4]
        droll = positions[1, 5]-positions[0, 5]

        self.velocity.x = x
        self.velocity.y = y
        self.velocity.z = z
        self.velocity.alpha_0 = yaw
        self.velocity.alpha_1 = pitch
        self.velocity.alpha_2 = roll
        self.velocity.dx = dx
        self.velocity.dy = dy
        self.velocity.dz = dz
        self.velocity.dalpha_0 = dyaw
        self.velocity.dalpha_1 = dpitch
        self.velocity.dalpha_2 = droll

        rof, hof, vof = mcode.optic_flow(self.scene, self.viewing_directions,
                                         self.velocity)

        testrof = [[-7.02912790e-58, -8.01829413e-51, 1.06910588e-50],
                   [0.00000000e+00, -2.11758237e-22, -4.23516474e-22],
                   [-2.25532711e-28, 1.69406589e-21, -1.69406589e-21]]
        testhof = [[-2.02066722e-17, -2.02066722e-17, -2.02066722e-17],
                   [-5.75929412e-03, -5.75929412e-03, -5.75929412e-03],
                   [-1.15168339e-02, -1.15168339e-02, -1.15168339e-02]]
        testvof = [[-1.14794370e-41, -1.44444746e-34, 1.92592994e-34],
                   [0.00000000e+00, -1.35525272e-20, -2.71050543e-20],
                   [-6.45841185e-27, 5.42101086e-20, 0.00000000e+00]]

        # print("sinel",sinel)
        # print("sin prop:", hof[:,0]/list(reversed(sinel)))
        # plt.plot(hof[:,0])
        # plt.plot(sinel)
        # plt.show()

        assert np.all(np.isclose(rof, testrof))
        assert np.all(np.isclose(hof, testhof))
        assert np.all(np.isclose(vof, testvof))

    def test_only_yaw_big(self):
        """
        this test checks for the correct response if for example
        the bee rotates only around the yaw axis keeping the other
        parameters (x,y,z,pitch,roll) constant.
        here the azimuthal optic flow should be zero and the
        elevational optic flow should be proportional with a
        constant factor to the cos of the elevation
        """
        # generate scene aka distance channel
        scene = np.random.random((180, 360, 4, 1))
        velocity = pd.Series(index=['x', 'y', 'z',
                                    'alpha_0', 'alpha_1', 'alpha_2',
                                    'dx', 'dy', 'dz',
                                    'dalpha_0', 'dalpha_1', 'dalpha_2'])

        elevation = np.arange(-np.pi/2, np.pi/2, 2*(np.pi/360))
        azimuth = np.arange(0, 2*np.pi, 2*(np.pi/360))

        viewing_directions = np.zeros((180, 2))
        viewing_directions[:, 0] = elevation
        viewing_directions[:, 1] = azimuth[0:180]

        positions = np.array([[-20, -20, 2.6, 1.57079633, 0, 0],
                              [-20, -20, 2.6, 1.90079633, 0, 0]])

        x = positions[0, 0]
        y = positions[0, 1]
        z = positions[0, 2]
        yaw = positions[0, 3]
        pitch = positions[0, 4]
        roll = positions[0, 5]

        dx = positions[1, 0]-positions[0, 0]
        dy = positions[1, 1]-positions[0, 1]
        dz = positions[1, 2]-positions[0, 2]
        dyaw = positions[1, 3]-positions[0, 3]
        dpitch = positions[1, 4]-positions[0, 4]
        droll = positions[1, 5]-positions[0, 5]

        velocity.x = x
        velocity.y = y
        velocity.z = z
        velocity.alpha_0 = yaw
        velocity.alpha_1 = pitch
        velocity.alpha_2 = roll
        velocity.dx = dx
        velocity.dy = dy
        velocity.dz = dz
        velocity.dalpha_0 = dyaw
        velocity.dalpha_1 = dpitch
        velocity.dalpha_2 = droll

        rof, hof, vof = mcode.optic_flow(scene, viewing_directions,
                                         velocity)
        cosel = np.cos(elevation)
        has_zeros = len(np.where(cosel == 0)) > 0
        factor = np.array([0])
        if not has_zeros:
            factor = np.array(hof[:, 0]/cosel)
        # print("sinel",sinel)
        # print("cos prop:", hof[:,0]/cosel)
        # fig, ax = plt.subplots()
        # ax.plot(elevation,hof[:,0],label='hof')
        # ax.plot(elevation,cosel,label='cos(el)')
        # plt.xlabel('elevation')
        # ax.legend()
        # plt.show()

        assert np.all(factor == factor[0])
        # assert np.all(np.isclose(hof, testhof))
        # assert np.all(np.isclose(vof, testvof))

    def test_only_pitch(self):
        """
        this test checks for the correct response if for example
        the bee rotates only around the pitch axis keeping the other
        parameters (x,y,z,yaw,roll) constant
        """
        positions = np.array([[-20, -20, 2.6, 1.57079633, 0, 0],
                              [-20, -20, 2.6, 1.57079633, 0.1, 0]])

        x = positions[0, 0]
        y = positions[0, 1]
        z = positions[0, 2]
        yaw = positions[0, 3]
        pitch = positions[0, 4]
        roll = positions[0, 5]

        dx = positions[1, 0] - positions[0, 0]
        dy = positions[1, 1] - positions[0, 1]
        dz = positions[1, 2] - positions[0, 2]
        dyaw = positions[1, 3] - positions[0, 3]
        dpitch = positions[1, 4] - positions[0, 4]
        droll = positions[1, 5] - positions[0, 5]

        self.velocity.x = x
        self.velocity.y = y
        self.velocity.z = z
        self.velocity.alpha_0 = yaw
        self.velocity.alpha_1 = pitch
        self.velocity.alpha_2 = roll
        self.velocity.dx = dx
        self.velocity.dy = dy
        self.velocity.dz = dz
        self.velocity.dalpha_0 = dyaw
        self.velocity.dalpha_1 = dpitch
        self.velocity.dalpha_2 = droll
        rof, hof, vof = mcode.optic_flow(self.scene, self.viewing_directions,
                                         self.velocity)

        testrof = [[0.00000000e+00, -1.20370622e-35, -2.40741243e-35],
                   [-8.07793567e-28, 0.00000000e+00, -6.77626358e-21],
                   [-1.61558713e-27, 0.00000000e+00, -2.71050543e-20]]
        testhof = [[-0.1, -0.09998477, -0.09993908],
                   [-0.09998477, -0.09996954, -0.09992386],
                   [-0.09993908, -0.09992386, -0.0998782]]
        testvof = [[-3.20510345e-10, -1.74524096e-03, -3.48994999e-03],
                   [-3.20510345e-10, -1.74524096e-03, -3.48994999e-03],
                   [-3.20510345e-10, -1.74524096e-03, -3.48994999e-03]]

        assert np.all(np.isclose(rof, testrof))
        assert np.all(np.isclose(hof, testhof))
        assert np.all(np.isclose(vof, testvof))

    def test_only_pitch_big(self):
        """
        this test checks for the correct response if for example
        the bee rotates only around the pitch axis keeping the other
        parameters (x,y,z,yaw,roll) constant.
        here the azimuthal optic flow should be zero and the
        elevational optic flow should be proportional with a
        constant factor to the cos of the elevation
        """
        scene = np.random.random((180, 360, 4, 1))

        velocity = pd.Series(index=['x', 'y', 'z',
                                    'alpha_0', 'alpha_1', 'alpha_2',
                                    'dx', 'dy', 'dz',
                                    'dalpha_0', 'dalpha_1', 'dalpha_2'])

        elevation = np.arange(-np.pi/2, np.pi/2, 2*(np.pi/360))

        azimuth = np.arange(0, 2*np.pi, 2*(np.pi/360))
        viewing_directions = np.zeros((180, 2))
        viewing_directions[:, 0] = elevation
        viewing_directions[:, 1] = azimuth[0:180]

        positions = np.array([[-20, -20, 2.6, 1.57079633, 0, 0],
                              [-20, -20, 2.6, 1.57079633, 0.1, 0]])

        x = positions[0, 0]
        y = positions[0, 1]
        z = positions[0, 2]
        yaw = positions[0, 3]
        pitch = positions[0, 4]
        roll = positions[0, 5]

        dx = positions[1, 0] - positions[0, 0]
        dy = positions[1, 1] - positions[0, 1]
        dz = positions[1, 2] - positions[0, 2]
        dyaw = positions[1, 3] - positions[0, 3]
        dpitch = positions[1, 4] - positions[0, 4]
        droll = positions[1, 5] - positions[0, 5]

        velocity.x = x
        velocity.y = y
        velocity.z = z
        velocity.alpha_0 = yaw
        velocity.alpha_1 = pitch
        velocity.alpha_2 = roll
        velocity.dx = dx
        velocity.dy = dy
        velocity.dz = dz
        velocity.dalpha_0 = dyaw
        velocity.dalpha_1 = dpitch
        velocity.dalpha_2 = droll

        rof, hof, vof = mcode.optic_flow(scene, viewing_directions,
                                         velocity)

        cosel = np.cos(elevation)
        has_zeros = len(np.where(cosel == 0)) > 0
        factor = np.array([0])
        if not has_zeros:
            factor = np.array(vof[0, :]/cosel)
        # print("cos prop:", vof[0, :]/cosel)
        # fig, ax = plt.subplots()
        # ax.plot(elevation, vof[0, :], label='hof')
        # ax.plot(elevation, cosel, label='cos(el)')
        # plt.xlabel('elevation')
        # ax.legend()
        # plt.show()

        assert np.all(factor == factor[0])
        # assert np.all(np.isclose(hof, testhof))
        # assert np.all(np.isclose(vof, testvof))

    def test_only_roll_big(self):
        """
        this test checks for the correct response if for example
        the bee rotates only around the roll axis keeping the other
        parameters (x,y,z,yaw,pitch) constant
        here the azimuthal optic flow should be zero and the
        elevational optic flow should be proportional with a
        constant factor to the cos of the elevation
        """
        scene = np.random.random((180, 360, 4, 1))

        velocity = pd.Series(index=['x', 'y', 'z',
                                    'alpha_0', 'alpha_1', 'alpha_2',
                                    'dx', 'dy', 'dz',
                                    'dalpha_0', 'dalpha_1', 'dalpha_2'])

        elevation = np.arange(-np.pi/2, np.pi/2, 2*(np.pi/360))

        azimuth = np.arange(0, 2*np.pi, 2*(np.pi/360))
        viewing_directions = np.zeros((180, 2))
        viewing_directions[:, 0] = elevation
        viewing_directions[:, 1] = azimuth[0:180]

        positions = np.array([[-20, -20, 2.6, 1.57079633, 0, 0],
                              [-20, -20, 2.6, 1.57079633, 0, 0.1]])

        x = positions[0, 0]
        y = positions[0, 1]
        z = positions[0, 2]
        yaw = positions[0, 3]
        pitch = positions[0, 4]
        roll = positions[0, 5]

        dx = positions[1, 0] - positions[0, 0]
        dy = positions[1, 1] - positions[0, 1]
        dz = positions[1, 2] - positions[0, 2]
        dyaw = positions[1, 3] - positions[0, 3]
        dpitch = positions[1, 4] - positions[0, 4]
        droll = positions[1, 5] - positions[0, 5]

        velocity.x = x
        velocity.y = y
        velocity.z = z
        velocity.alpha_0 = yaw
        velocity.alpha_1 = pitch
        velocity.alpha_2 = roll
        velocity.dx = dx
        velocity.dy = dy
        velocity.dz = dz
        velocity.dalpha_0 = dyaw
        velocity.dalpha_1 = dpitch
        velocity.dalpha_2 = droll

        rof, hof, vof = mcode.optic_flow(scene, viewing_directions,
                                         velocity)
        cosel = np.cos(elevation)
        sinel = np.sin(elevation)
        has_zeros = cosel[np.where(cosel == 0)] = 1
        factor = np.array(hof[0, :]/cosel)
        factor[has_zeros] = factor[1]
        has_zerossin = sinel[np.where(sinel == 0)] = 1
        factorsin = np.array(vof[0, :]/sinel)
        factorsin[has_zerossin] = factorsin[0]
        # print("cos prop:", vof[0, :]/cosel)
        # fig, ax = plt.subplots()
        # ax.plot(elevation, hof[0, :], label='hof')
        # ax.plot(elevation, vof[0, :], label='vof')
        # ax.plot(elevation, cosel, label='cos(el)')
        # ax.plot(elevation, sinel, label='sin(el)')
        # plt.xlabel('elevation')
        # ax.legend()
        # plt.show()

        # fig, ax = plt.subplots()
        # ax.plot(azimuth[0:180], hof[:, 0], label='hof')
        # ax.plot(azimuth[0:180], vof[:, 0], label='vof')
        # plt.xlabel('azimuth')
        # ax.legend()
        # plt.show()

        for i in range(1, len(factor)):
            assert np.isclose(factor[1], factor[i])
            if i == 90:
                continue
            assert np.isclose(factorsin[0], factorsin[i])
            # print(i)
        # assert np.all(np.isclose(hof, testhof))
        # assert np.all(np.isclose(vof, testvof))

    def test_only_roll(self):
        """
        this test checks for the correct response if for example
        the bee rotates only around the pitch axis keeping the other
        parameters (x,y,z,yaw,roll) constant
        """
        positions = np.array([[-20, -20, 2.6, 1.57079633, 0.1, 0.1],
                              [-20, -20, 2.6, 1.57079633, 0.1, 0.2]])

        x = positions[0, 0]
        y = positions[0, 1]
        z = positions[0, 2]
        yaw = positions[0, 3]
        pitch = positions[0, 4]
        roll = positions[0, 5]

        dx = positions[1, 0] - positions[0, 0]
        dy = positions[1, 1] - positions[0, 1]
        dz = positions[1, 2] - positions[0, 2]
        dyaw = positions[1, 3] - positions[0, 3]
        dpitch = positions[1, 4] - positions[0, 4]
        droll = positions[1, 5] - positions[0, 5]

        self.velocity.x = x
        self.velocity.y = y
        self.velocity.z = z
        self.velocity.alpha_0 = yaw
        self.velocity.alpha_1 = pitch
        self.velocity.alpha_2 = roll
        self.velocity.dx = dx
        self.velocity.dy = dy
        self.velocity.dz = dz
        self.velocity.dalpha_0 = dyaw
        self.velocity.dalpha_1 = dpitch
        self.velocity.dalpha_2 = droll
        rof, hof, vof = mcode.optic_flow(self.scene, self.viewing_directions,
                                         self.velocity)

        testrof = [[-4.88735544e-05, -4.88735544e-05, -4.88735544e-05],
                   [-4.91907871e-05, -4.94889779e-05, -4.97869621e-05],
                   [-4.94479358e-05, -5.00455194e-05, -5.06426694e-05]]
        testhof = [[1.48878092e-05, 1.75149414e-03, 3.48756694e-03],
                   [-1.61062585e-04, 1.57529538e-03, 3.31112146e-03],
                   [-3.36963918e-04,  1.39861677e-03, 3.13366737e-03]]
        testvof = [[-0.09950539, -0.09948998, -0.09944426],
                   [-0.09950368, -0.0994883, -0.09944262],
                   [-0.09950195, -0.09948661, -0.09944095]]

        assert np.all(np.isclose(rof, testrof))
        assert np.all(np.isclose(hof, testhof))
        assert np.all(np.isclose(vof, testvof))


if __name__ == '__main__':
    unittest.main()
